"""
randcraft.rvs
"""

from randcraft.rvs.anonymous import AnonymousRV
from randcraft.rvs.base import RV
from randcraft.rvs.discrete import DiracDeltaRV, DiscreteRV
from randcraft.rvs.gaussian_kde import GaussianKdeRV
from randcraft.rvs.mixture import MixtureRV
from randcraft.rvs.multi import MultiRV
from randcraft.rvs.scipy_pdf import SciRV

__all__ = [
    "RV",
    "DiscreteRV",
    "DiracDeltaRV",
    "AnonymousRV",
    "MixtureRV",
    "SciRV",
    "GaussianKdeRV",
    "MultiRV",
]



import numpy as np

from randcraft.observations import make_gaussian_kde
from tests.base_test_case import BaseTestCase


class TestObservations(BaseTestCase):
    def test_kde(self) -> None:
        observations_a = np.array([1.0, 2.0, 3.0, 4.0, 5.0]) * 1
        rv_a = make_gaussian_kde(observations=observations_a, bw_method=0.5)

        observations_b = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0])
        rv_b = make_gaussian_kde(observations=observations_b)
        rv_joined = rv_a + rv_b

        self.assertAlmostEqual(rv_a.get_mean(), 3.0)
        self.assertAlmostEqual(rv_b.get_mean(), 0.5)
        self.assertAlmostEqual(rv_joined.get_mean(), 3.0 + 0.5)

        self.assertAlmostEqual(rv_a.get_variance(), float(np.var(observations_a)) * 2.25)



from collections.abc import Callable

import numpy as np

from randcraft.observations import make_gaussian_kde
from randcraft.random_variable import RandomVariable
from randcraft.rvs.anonymous import AnonymousRV
from randcraft.rvs.continuous import ContinuousRV


def make_anonymous_continuous_rv_from_sampler(sampler: Callable[[int], np.ndarray], n_observations: int = 500) -> RandomVariable:
    """
    Creates a random variable using a continuous sampling function.
    Statistics, cdf, pdf etc are estimated by making a kde using random observations from the sampler.

    Args:
        sampler (Callable[[int], np.ndarray]): A function that takes an integer and returns a numpy array of observations.
        n_observations (int, optional): The number of observations to draw from the sampler. Defaults to 10000.

    Returns:
        RandomVariable: A random variable containing the sampler and statistics estimated using a kde.
    """
    observations = sampler(n_observations)
    assert len(np.unique(observations)) / n_observations > 0.9, "It appears that the provided sampler is not continuous"
    ref_rv: ContinuousRV = make_gaussian_kde(observations=observations)._rv  # type: ignore
    return RandomVariable(rv=AnonymousRV(sampler=sampler, ref_rv=ref_rv))



from randcraft.constructors import make_beta, make_coin_flip, make_die_roll, make_dirac, make_discrete, make_normal, make_uniform
from randcraft.random_variable import RandomVariable
from randcraft.rvs import DiracDeltaRV, DiscreteRV
from tests.base_test_case import BaseTestCase


class TestCombiningRvs(BaseTestCase):
    def test_scaled_rv(self) -> None:
        a = 2.0
        b = 5.0

        rv = make_beta(a=a, b=b)

        scaled_rv = rv.scale(factor=1)
        self.assert_stats_are_close(a=scaled_rv.statistics, b=rv.statistics)

        offset_rv = scaled_rv - 1
        self.assertAlmostEqual(offset_rv.get_mean(), rv.get_mean() - 1)

        negative_rv = rv.scale(factor=-1)
        new_rv = negative_rv - 2
        self.assertAlmostEqual(new_rv.get_mean(), rv.get_mean() * -1 - 2)
        self.assertAlmostEqual(new_rv.get_variance(), rv.get_variance())

    def test_combining_different_types_of_rvs(self) -> None:
        rv1 = make_normal(mean=10, std_dev=2)
        rv2 = make_uniform(low=80, high=120)

        partial_combined_rv = rv1 + rv2
        self.assertIsInstance(partial_combined_rv, RandomVariable)
        self.assertAlmostEqual(partial_combined_rv.get_mean(exact=True), 10 + 100)
        self.assertAlmostEqual(partial_combined_rv.get_variance(exact=True), rv1.get_variance() + rv2.get_variance())

    def test_pdf_combination(self) -> None:
        rv1 = make_uniform(low=0, high=10)
        rv2 = make_dirac(value=10)
        combined = rv1 + rv2
        self.assertEqual(combined.cdf(value=10), 0.0)
        self.assertEqual(combined.cdf(value=20), 1.0)

    def test_normal_known_convolutions(self) -> None:
        rv1 = make_normal(mean=1, std_dev=4)
        rv2 = make_normal(mean=2, std_dev=1)
        dirac1 = make_dirac(value=1)
        combined = rv1 + rv2

        self.assertIsInstance(combined, RandomVariable)
        self.assertAlmostEqual(combined.get_mean(exact=True), 3)
        self.assertAlmostEqual(combined.get_variance(exact=True), 17)

        offset = combined + dirac1
        self.assertIsInstance(offset, RandomVariable)
        self.assertAlmostEqual(offset.get_mean(exact=True), 4)
        self.assertAlmostEqual(offset.get_variance(exact=True), 17)

    def test_normal_arithmetic(self) -> None:
        rv1 = make_normal(mean=5, std_dev=1)
        rv2 = make_normal(mean=3, std_dev=1)

        # Test addition
        rv_add = rv1 + rv2
        self.assertIsInstance(rv_add, RandomVariable)
        self.assertAlmostEqual(rv_add.get_mean(), 5 + 3)
        self.assertAlmostEqual(rv_add.get_variance(), rv1.get_variance() + rv2.get_variance())

        # Test subtraction
        rv_sub = rv1 - rv2
        self.assertIsInstance(rv_sub, RandomVariable)
        self.assertAlmostEqual(rv_sub.get_mean(), 5 - 3)
        self.assertAlmostEqual(rv_sub.get_variance(), rv1.get_variance() + rv2.get_variance())

        # Test scaling
        rv_neg = rv1.scale(-1)
        self.assertIsInstance(rv_neg, RandomVariable)
        self.assertAlmostEqual(rv_neg.get_mean(), -5)
        self.assertAlmostEqual(rv_neg.get_variance(), rv1.get_variance())

        rv_div = rv1.scale(1 / 2)
        self.assertIsInstance(rv_div, RandomVariable)
        self.assertAlmostEqual(rv_div.get_mean(), 5 / 2)
        self.assertAlmostEqual(rv_div.get_variance(), rv1.get_variance() / 4)

        # Test scale by zero
        rv_zero_scale = rv1.scale(0)
        self.assertIsInstance(rv_zero_scale, RandomVariable)
        self.assertAlmostEqual(rv_zero_scale.get_mean(), 0.0)
        self.assertAlmostEqual(rv_zero_scale.get_variance(), 0.0)
        self.assertIsInstance(rv_zero_scale._rv, DiracDeltaRV)

    def test_uniform_known_convolutions(self) -> None:
        rv1 = make_uniform(low=0, high=10)
        dirac1 = make_dirac(value=1)
        offset = rv1 + dirac1

        self.assertIsInstance(offset, RandomVariable)
        self.assertAlmostEqual(offset.get_mean(exact=True), 6)
        self.assertAlmostEqual(offset.get_variance(exact=True), rv1.get_variance())

    def test_uniform_multiplications(self) -> None:
        rv1 = make_uniform(low=0, high=10)

        double_rv1 = rv1.scale(2)
        self.assertIsInstance(double_rv1, RandomVariable)
        self.assertAlmostEqual(double_rv1.get_mean(exact=True), 10)
        self.assertAlmostEqual(double_rv1.get_variance(exact=True), rv1.get_variance() * 4)

        negative_rv1 = rv1.scale(-1)
        self.assertIsInstance(negative_rv1, RandomVariable)
        self.assertAlmostEqual(negative_rv1.get_mean(exact=True), -5)
        self.assertAlmostEqual(negative_rv1.get_variance(exact=True), rv1.get_variance())

    def test_discrete_convolution(self) -> None:
        dice = make_discrete(values=[1, 2, 3, 4, 5, 6])
        two_dice = dice + dice
        self.assertIsInstance(two_dice, RandomVariable)
        self.assertIsInstance(two_dice._rv, DiscreteRV)
        self.assertAlmostEqual(two_dice.get_mean(exact=True), 7.0)

        one = make_dirac(value=1)
        two_dice_and_one = two_dice + one
        self.assertIsInstance(two_dice_and_one, RandomVariable)
        self.assertIsInstance(two_dice_and_one._rv, DiscreteRV)
        self.assertAlmostEqual(two_dice_and_one.get_mean(exact=True), 8.0)

    def test_discrete_and_continuous_convolution(self) -> None:
        coin_flip = make_coin_flip()
        norm = make_normal(mean=0, std_dev=0.2)
        combined = coin_flip + norm
        print(combined)
        self.assertEqual(combined.get_variance(), coin_flip.get_variance() + norm.get_variance())
        self.assertEqual(combined.get_mean(), 0.5)

    def test_three_dice(self) -> None:
        dice = make_die_roll(sides=6)
        three_dice = dice + dice + dice
        result = three_dice.cdf(10.0)
        self.assertEqual(result, 0.5)

    def test_complex_convolution(self) -> None:
        rv_base = make_uniform(low=-1, high=1)
        rv = rv_base + rv_base
        discrete = make_discrete(values=[0, 6])
        new_rv = rv + discrete
        self.assertAlmostEqual(new_rv.cdf(-2.0), 0.0, places=3)
        self.assertAlmostEqual(new_rv.cdf(0.0), 0.25, places=3)
        self.assertAlmostEqual(new_rv.cdf(3.0), 0.5, places=3)
        self.assertAlmostEqual(new_rv.cdf(6.0), 0.75, places=3)
        self.assertAlmostEqual(new_rv.cdf(8.0), 1.0, places=3)



from randcraft.random_variable import RandomVariable
from randcraft.rvs import (
    DiracDeltaRV,
    MixtureRV,
)


def mix_rvs(rvs: list[RandomVariable], probabilities: list[float] | None = None) -> RandomVariable:
    pdfs = [rv._rv for rv in rvs]
    return RandomVariable(rv=MixtureRV(pdfs=pdfs, probabilities=probabilities))  # type: ignore


def add_special_event_to_rv(rv: RandomVariable, value: float, chance: float) -> RandomVariable:
    assert 0 <= chance <= 1.0, "Value must be between 0 and 1"
    dirac_rv = RandomVariable(DiracDeltaRV(value=value))
    return mix_rvs(rvs=[rv, dirac_rv], probabilities=[1.0 - chance, chance])



from collections.abc import Sequence
from functools import cached_property

import numpy as np

from randcraft.models import ProbabilityDensityFunction, ProbabilityMassFunction, Statistics, Uncertainty, join_uncertainties, sum_uncertain_floats
from randcraft.rvs.base import RV, CdfEstimator
from randcraft.rvs.continuous import ContinuousRV
from randcraft.rvs.discrete import DiscreteRV


class MixtureRV(RV):
    def __init__(
        self,
        pdfs: Sequence[ContinuousRV | DiscreteRV],
        probabilities: list[float] | None = None,
    ) -> None:
        # TODO Add support for mixture of mixtures by flattening the input
        if probabilities is None:
            probabilities = [1.0 / len(pdfs)] * len(pdfs)
        self._pdfs = pdfs
        self._probabilities = probabilities
        self._validate()

    def _validate(self) -> None:
        pdfs = self._pdfs
        probabilities = self._probabilities
        assert len(pdfs) > 0, "At least one pdf is required"
        assert all(isinstance(pdf, ContinuousRV | DiscreteRV) for pdf in pdfs), f"All pdfs must be instances of {ContinuousRV | DiscreteRV}"
        assert len(pdfs) == len(probabilities), "Number of PDFs must match number of weights"
        assert all(w > 0 for w in probabilities), "All weights must be positive"
        total_weight = sum(probabilities)
        assert abs(total_weight - 1.0) < 1e-9, "Weights must sum to 1"

    @property
    def short_name(self) -> str:
        return "mixture"

    @cached_property
    def statistics(self) -> Statistics:
        mean = sum_uncertain_floats(pdf.statistics.mean * weight for pdf, weight in zip(self.pdfs, self.probabilities))
        second_moment = sum_uncertain_floats(pdf.statistics.moments[1] * weight for pdf, weight in zip(self.pdfs, self.probabilities))
        # TODO calculate more moments

        extreme_values = [pdf.statistics.min_value for pdf in self.pdfs] + [pdf.statistics.max_value for pdf in self.pdfs]
        min_value = min(extreme_values, key=lambda x: x.value)
        max_value = max(extreme_values, key=lambda x: x.value)

        return Statistics(
            moments=[mean, second_moment],
            support=(min_value, max_value),
        )

    @property
    def pdfs(self) -> Sequence[ContinuousRV | DiscreteRV]:
        return self._pdfs

    @property
    def probabilities(self) -> list[float]:
        return self._probabilities

    def _get_discrete_points(self) -> np.ndarray:
        return np.sort(np.unique(np.concatenate([pdf._get_discrete_points() for pdf in self.pdfs])))

    def scale(self, x: float) -> "MixtureRV":
        x = float(x)
        return MixtureRV(pdfs=[pdf.scale(x) for pdf in self.pdfs], probabilities=self.probabilities)

    def add_constant(self, x: float) -> "MixtureRV":
        return MixtureRV(pdfs=[pdf.add_constant(x) for pdf in self.pdfs], probabilities=self.probabilities)

    def sample_numpy(self, n: int) -> np.ndarray:
        rng = np.random.default_rng()
        pdf_choices = rng.choice(len(self.pdfs), size=n, p=self.probabilities)

        samples = np.zeros(n)
        for i in range(len(self.pdfs)):
            # Get the number of samples to draw from this PDF
            num_samples = np.sum(pdf_choices == i)
            if num_samples > 0:
                # Draw samples from the PDF and add them to the result
                samples[pdf_choices == i] = self.pdfs[i].sample_numpy(num_samples)
        return samples

    @cached_property
    def _cdf_estimator(self) -> CdfEstimator:
        return CdfEstimator(rv=self)

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        def func(cdfs: list[np.ndarray]) -> np.ndarray:
            cdf = np.sum([cdf * p for cdf, p in zip(cdfs, self.probabilities)], axis=0)
            cdf[x > self.statistics.max_value.value] = 1.0
            cdf[x <= self.statistics.min_value.value] = 0.0
            return cdf

        cdfs = [pdf.cdf(x) for pdf in self.pdfs]
        return join_uncertainties(uncertainties=cdfs, func=func)  # type: ignore

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._cdf_estimator.ppf(x)

    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction | None:
        pdf_prob_pairs = [(pdf, p) for pdf, p in zip(self.pdfs, self.probabilities) if isinstance(pdf, ContinuousRV)]
        if not len(pdf_prob_pairs):
            return None
        y = np.zeros_like(x)
        for pdf, p in pdf_prob_pairs:
            cont_pdf = pdf.calculate_pdf(x)
            assert cont_pdf is not None
            y += cont_pdf.y * p
        return ProbabilityDensityFunction(x=x, y=y)

    def calculate_pmf(self) -> ProbabilityMassFunction | None:
        pdf_prob_pairs = [(pdf, p) for pdf, p in zip(self.pdfs, self.probabilities) if isinstance(pdf, DiscreteRV)]
        if not len(pdf_prob_pairs):
            return None

        value_probs: dict[float, float] = {}
        for pdf, p in pdf_prob_pairs:
            disc_pdf = pdf.calculate_pmf()
            assert disc_pdf is not None
            for v, prob in zip(disc_pdf.x, disc_pdf.y):
                if v in value_probs:
                    value_probs[v] += prob * p
                else:
                    value_probs[v] = prob * p
        values, probs = zip(*sorted(value_probs.items()))
        y = np.array(probs)
        x = np.array(values)
        return ProbabilityMassFunction(x=x, y=y)

    def copy(self) -> "MixtureRV":
        return MixtureRV(pdfs=self.pdfs, probabilities=self.probabilities)



from randcraft import make_dirac, make_normal
from randcraft.misc import add_special_event_to_rv, mix_rvs
from randcraft.random_variable import RandomVariable
from randcraft.rvs import (
    MixtureRV,
)
from tests.base_test_case import BaseTestCase


class TestMisc(BaseTestCase):
    def test_mixture_rv(self) -> None:
        rv_a = make_dirac(1.0)
        rv_b = make_normal(mean=2.0, std_dev=2.0)
        mixture = mix_rvs([rv_a, rv_b])

        self.assertIsInstance(mixture, RandomVariable)
        self.assertIsInstance(mixture._rv, MixtureRV)
        self.assertAlmostEqual(mixture.get_mean(exact=True), (1.0 + 2.0) / 2)
        # TODO Test variance of mixture, test using different weights

    def test_add_special_event(self) -> None:
        rv = make_dirac(value=1.0)
        new_rv = add_special_event_to_rv(rv=rv, value=2.0, chance=0.5)

        self.assertIsInstance(new_rv, RandomVariable)
        self.assertAlmostEqual(new_rv.get_mean(), 1.5)



import numpy as np
from scipy.stats import beta, gamma, norm, uniform
from scipy.stats._distn_infrastructure import rv_continuous

from randcraft.random_variable import RandomVariable
from randcraft.rvs import (
    DiracDeltaRV,
    DiscreteRV,
    SciRV,
)
from randcraft.utils import clean_1d_array

__all__ = [
    "make_discrete",
    "make_dirac",
    "make_coin_flip",
    "make_die_roll",
    "make_scipy",
    "make_normal",
    "make_uniform",
    "make_beta",
    "make_gamma",
]


# Discrete
def make_discrete(values: list[float] | list[int] | np.ndarray, probabilities: list[float] | np.ndarray | None = None) -> RandomVariable:
    # If probabilities are not provided, equal probabilities are assumed
    values = clean_1d_array(values)
    if probabilities is not None:
        probabilities = clean_1d_array(probabilities)
    return RandomVariable(rv=DiscreteRV(values=values, probabilities=probabilities))


def make_dirac(value: float | int) -> RandomVariable:
    value = float(value)
    return RandomVariable(rv=DiracDeltaRV(value=value))


def make_coin_flip() -> RandomVariable:
    return make_discrete(values=[0, 1], probabilities=[0.5, 0.5])


def make_die_roll(sides: int = 6) -> RandomVariable:
    values = list(range(1, sides + 1))
    probabilities = [1 / sides] * sides
    return make_discrete(values=values, probabilities=probabilities)


# Scipy
def make_scipy(scipy_rv: rv_continuous, *args: float | int, **kwargs: float | int) -> RandomVariable:
    return RandomVariable(rv=SciRV(scipy_rv, *args, **kwargs))


# Helpers for common scipy distributions
def make_normal(mean: float | int, std_dev: float | int) -> RandomVariable:
    mean = float(mean)
    std_dev = float(std_dev)
    return make_scipy(scipy_rv=norm, loc=mean, scale=std_dev)


def make_uniform(low: float | int, high: float | int) -> RandomVariable:
    low = float(low)
    high = float(high)
    return make_scipy(scipy_rv=uniform, loc=low, scale=high - low)


def make_beta(a: float | int, b: float | int) -> RandomVariable:
    a = float(a)
    b = float(b)
    return make_scipy(scipy_rv=beta, a=a, b=b)


def make_gamma(a: float | int, scale: float | int) -> RandomVariable:
    a = float(a)
    scale = float(scale)
    return make_scipy(scipy_rv=gamma, a=a, scale=scale)



from unittest import TestCase

from randcraft.models import Statistics


class BaseTestCase(TestCase):
    def assert_stats_are_close(self, a: Statistics, b: Statistics, places: int = 5) -> None:
        moments_to_check = 2
        for k in range(moments_to_check):
            self.assertAlmostEqual(a.moments[k].value, b.moments[k].value, places=places)
        self.assertAlmostEqual(a.min_value.value, b.min_value.value, places=places)
        self.assertAlmostEqual(a.max_value.value, b.max_value.value, places=places)



from collections.abc import Sequence

from scipy.stats import norm

from randcraft.models import AlgebraicFunction
from randcraft.rvs.continuous import ContinuousRV, ScaledRV
from randcraft.rvs.discrete import DiracDeltaRV, DiscreteRV
from randcraft.rvs.multi import MultiRV
from randcraft.rvs.scipy_pdf import SciRV


class PdfConvolver:
    @classmethod
    def convolve_pdfs(cls, pdfs: Sequence[DiscreteRV | ContinuousRV]) -> DiscreteRV | ContinuousRV:
        # TODO How to deal with mixtures here?

        assert len(pdfs) >= 2, "At least two PDFs are required for combination."
        if not all(isinstance(pdf, ContinuousRV | DiscreteRV) for pdf in pdfs):
            types = {pdf.__class__.__name__ for pdf in pdfs}
            raise TypeError(f"All PDFs must be instances of {ContinuousRV | DiscreteRV}, got: {types}")

        if all(rv.short_name == "normal" for rv in pdfs):
            return cls.convolve_normals(pdfs=pdfs)  # type: ignore

        continuous_pdfs, discrete_pdfs = cls.unpack_pdfs(pdfs=pdfs)

        if len(discrete_pdfs) >= 2:
            # First we convolve together all discrete pdfs
            reduced_discrete = cls.convolve_discretes(pdfs=discrete_pdfs)

            if not len(continuous_pdfs):
                return reduced_discrete

            discrete_pdfs = [reduced_discrete]

        discrete_pdf = discrete_pdfs[0] if len(discrete_pdfs) else None
        if discrete_pdf is not None:
            if len(continuous_pdfs) == 1 and isinstance(discrete_pdf, DiracDeltaRV):
                return continuous_pdfs[0].add_constant(discrete_pdf.value)
            return MultiRV(continuous_pdfs=continuous_pdfs, discrete_pdf=discrete_pdf)

        if len(continuous_pdfs) == 1:
            return continuous_pdfs[0]
        return MultiRV(continuous_pdfs=continuous_pdfs)

    @classmethod
    def unpack_pdfs(cls, pdfs: Sequence[DiscreteRV | ContinuousRV]) -> tuple[list[ContinuousRV], list[DiscreteRV]]:
        continuous_pdfs: list[ContinuousRV] = []
        discrete_pdfs: list[DiscreteRV] = []

        for pdf in pdfs:
            if isinstance(pdf, ScaledRV):
                af = pdf.algebraic_function
                no_offset_continuous = ScaledRV(inner=pdf.inner, algebraic_function=AlgebraicFunction(scale=af.scale, offset=0.0))
                continuous_pdfs.append(no_offset_continuous)
                if af.offset != 0.0:
                    offset_distribution = DiracDeltaRV(value=af.offset)
                    discrete_pdfs.append(offset_distribution)
            elif isinstance(pdf, MultiRV):
                inner_continuous, inner_discrete = cls.unpack_pdfs(pdf.pdfs)
                continuous_pdfs.extend(inner_continuous)
                discrete_pdfs.extend(inner_discrete)
            elif isinstance(pdf, ContinuousRV):
                continuous_pdfs.append(pdf)
            elif isinstance(pdf, DiscreteRV):
                discrete_pdfs.append(pdf)
            else:
                raise TypeError(f"Unsupported PDF type: {type(pdf)}")

        return continuous_pdfs, discrete_pdfs

    @classmethod
    def convolve_discretes(cls, pdfs: list[DiscreteRV]) -> DiscreteRV:
        assert all(isinstance(pdf, DiscreteRV) for pdf in pdfs)

        def to_dict(x: DiscreteRV) -> dict[float, float]:
            return {k: v for k, v in zip(x.values, x.probabilities)}

        def from_dict(x: dict[float, float]) -> DiscreteRV:
            values = list(x.keys())
            probabilities = list(x.values())
            return DiscreteRV(values=values, probabilities=probabilities)

        def convolve_two(x: dict[float, float], y: dict[float, float]) -> dict[float, float]:
            output_dict: dict[float, float] = {}
            for x1, p1 in x.items():
                for x2, p2 in y.items():
                    x3 = x1 + x2
                    p3 = p1 * p2
                    output_dict[x3] = output_dict.get(x3, 0.0) + p3
            return output_dict

        def convolve_iteratively(dicts: list[dict[float, float]]) -> dict[float, float]:
            if len(dicts) == 1:
                return dicts[0]
            first_two_joined = convolve_two(dicts[0], dicts[1])
            others = dicts[2:]
            if not len(others):
                return first_two_joined
            new_dicts = [first_two_joined, *others]
            return convolve_iteratively(new_dicts)

        dict_list = [to_dict(pdf) for pdf in pdfs]
        result_dict = convolve_iteratively(dict_list)
        result_pdf = from_dict(result_dict)
        if len(result_pdf.values) == 1:
            return DiracDeltaRV(value=result_pdf.values[0])
        return result_pdf

    @classmethod
    def convolve_normals(cls, pdfs: list[SciRV]) -> SciRV:
        # Equivalent to adding independent normal random variables
        if not pdfs:
            raise ValueError("No PDFs provided for combination.")

        for pdf in pdfs:
            assert isinstance(pdf, SciRV)
            assert pdf.short_name == "normal", f"Expected normal distribution, got {pdf.short_name}"

        new_mean = sum([pdf.mean for pdf in pdfs])
        new_variance = sum([pdf.variance for pdf in pdfs])
        return SciRV(norm, loc=new_mean, scale=new_variance**0.5)



from unittest import TestCase

from randcraft import make_discrete, make_normal, make_uniform
from randcraft.constructors import make_beta
from randcraft.misc import mix_rvs


class TestPlotting(TestCase):
    def test_plotting_convolved_uniform(self) -> None:
        rv_base = make_uniform(low=-1, high=1)
        rv = rv_base + rv_base
        rv.plot()

        discrete = make_discrete(values=[0, 6])
        new_rv = rv + discrete
        new_rv.plot()

    def test_plotting_mixed(self) -> None:
        rv1 = make_normal(mean=0, std_dev=1)
        rv2 = make_uniform(low=-1, high=1)
        combined = rv1 + rv2
        discrete = make_discrete(values=[1, 2, 3])
        mixed = mix_rvs([rv1, rv2, combined, discrete])
        mixed.plot()

    def test_scaled_plotting(self) -> None:
        a = 2.0
        b = 5.0

        rv = make_beta(a=a, b=b)
        rv.plot()

        new_rv = rv.scale(-1) - 2
        new_rv.plot()



from functools import cached_property

import numpy as np
from scipy.stats import gaussian_kde

from randcraft.models import ProbabilityDensityFunction, Statistics, Uncertainty, certainly
from randcraft.rvs.base import CdfEstimator
from randcraft.rvs.continuous import ContinuousRV
from randcraft.rvs.discrete import DiscreteRV


class GaussianKdeRV(ContinuousRV):
    def __init__(self, discrete: DiscreteRV, kde: gaussian_kde) -> None:
        self._discrete = discrete
        self._kde = kde

    @property
    def short_name(self) -> str:
        return "gaussian_kde"

    @cached_property
    def statistics(self) -> Statistics:
        mean = self._discrete.statistics.mean.make_uncertain()
        scale_factor: float = (1.0 + self._kde.factor) ** 2  # type: ignore
        variance = self._discrete.statistics.variance * scale_factor
        second_moment = mean**2 + variance.make_uncertain()
        # TODO calculate more moments

        min_value = certainly(-1 * float("inf"))
        max_value = certainly(float("inf"))

        return Statistics(
            moments=[mean, second_moment],
            support=(min_value, max_value),
        )

    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction:
        y = self._kde.evaluate(x)
        return ProbabilityDensityFunction(x=x, y=y)

    @cached_property
    def _cdf_estimator(self) -> CdfEstimator:
        return CdfEstimator(rv=self)

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._cdf_estimator.cdf(x)

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._cdf_estimator.ppf(x)

    def sample_numpy(self, n: int) -> np.ndarray:
        return self._kde.resample(n)[0]

    def copy(self) -> "GaussianKdeRV":
        return GaussianKdeRV(discrete=self._discrete, kde=self._kde)



from collections.abc import Callable, Iterable
from dataclasses import dataclass
from functools import cached_property
from typing import Literal, Protocol, TypeVar, runtime_checkable

import numpy as np

T = TypeVar("T")
U = TypeVar("U")


class SupportsFloatMul(Protocol):
    def __mul__(self: T, x: float, /) -> T: ...
    def __rmul__(self: T, x: float, /) -> T: ...


T_SFM = TypeVar("T_SFM", bound=SupportsFloatMul)


@dataclass(frozen=True)
class Uncertainty[T]:
    value: T
    is_certain: bool

    def __post_init__(self) -> None:
        assert not isinstance(self.value, Uncertainty), "Nested Uncertainty instances are not allowed"

    def __str__(self) -> str:
        if self.is_certain:
            return f"{self.value} (certain)"
        return f"{self.value} (uncertain)"

    def __repr__(self) -> str:
        return str(self)

    def __mul__(self, other: "Uncertainty[T]" | T | float | int) -> "Uncertainty[T]":
        return self._combine(other=other, func=lambda x, y: x * y)  # type: ignore

    def __add__(self, other: "Uncertainty[T]" | T | float | int) -> "Uncertainty[T]":
        return self._combine(other=other, func=lambda x, y: x + y)  # type: ignore

    def __sub__(self, other: "Uncertainty[T]" | T | float | int) -> "Uncertainty[T]":
        return self._combine(other=other, func=lambda x, y: x - y)  # type: ignore

    def __pow__(self, other: int | float) -> "Uncertainty[T]":
        assert not isinstance(other, Uncertainty), "Exponentiation with uncertain exponent is not supported"
        return self.apply(func=lambda x: x**other)  # type: ignore

    def apply(self, func: Callable[[T], T]) -> "Uncertainty[T]":
        return Uncertainty(value=func(self.value), is_certain=self.is_certain)

    def get(self, name: str = "value", certain: bool = False) -> T:
        if certain and not self.is_certain:
            raise NotImplementedError(f"{name} is uncertain")
        return self.value

    def make_uncertain(self) -> "Uncertainty[T]":
        return Uncertainty(value=self.value, is_certain=False)

    def _combine(self, other: "Uncertainty[T]" | T, func: Callable[[T, T], T]) -> "Uncertainty[T]":
        if not isinstance(other, Uncertainty):
            other = Uncertainty(value=other, is_certain=True)
        return Uncertainty(value=func(self.value, other.value), is_certain=self.is_certain and other.is_certain)


def certainly[T](x: T) -> Uncertainty[T]:
    return Uncertainty(value=x, is_certain=True)


def maybe[T](x: T) -> Uncertainty[T]:
    return Uncertainty(value=x, is_certain=False)


def sum_uncertains[T_SFM: SupportsFloatMul](uncertainties: Iterable[Uncertainty[T_SFM]]) -> Uncertainty[T_SFM]:
    values = list(uncertainties)
    assert len(values) > 0, "At least one uncertainty is required to sum"
    first_value = values[0]
    return sum(values, start=first_value * 0.0)


def sum_uncertain_floats(uncertainties: Iterable[Uncertainty[float]]) -> Uncertainty[float]:
    values = list(uncertainties)
    assert len(values) > 0, "At least one uncertainty is required to sum"
    return sum(values, certainly(0.0))


def join_uncertainties(uncertainties: Iterable[Uncertainty[T]], func: Callable[[Iterable[T]], U]) -> Uncertainty[U]:
    is_certain = all(u.is_certain for u in uncertainties)
    values = [u.value for u in uncertainties]
    return Uncertainty(value=func(values), is_certain=is_certain)


def sort_uncertainties(uncertainties: Iterable[Uncertainty[float]]) -> list[Uncertainty[float]]:
    return sorted(uncertainties, key=lambda x: x.value)


@dataclass(frozen=True)
class Statistics:
    moments: list[Uncertainty[float]]
    support: tuple[Uncertainty[float], Uncertainty[float]]

    def __post_init__(self) -> None:
        for m in self.moments:
            assert isinstance(m, Uncertainty), "All moments must be of type Uncertainty"
        for s in self.support:
            assert isinstance(s, Uncertainty), "All support values must be of type Uncertainty"
        assert len(self.moments) >= 2, f"At least 2 moments must be defined. Got {len(self.moments)}"

    @cached_property
    def central_moments(self) -> list[Uncertainty[float]]:
        def calculate_nth_central_moment(n: int) -> Uncertainty[float]:
            mean = self.mean
            non_central_moment = self.moments[n]
            return non_central_moment - mean ** (n + 1)

        return [calculate_nth_central_moment(n) for n, _ in enumerate(self.moments)]

    @property
    def mean(self) -> Uncertainty[float]:
        return self.moments[0]

    @property
    def variance(self) -> Uncertainty[float]:
        return self.central_moments[1]

    @cached_property
    def std_dev(self) -> Uncertainty[float]:
        return self.variance.apply(lambda x: x**0.5)

    @property
    def min_value(self) -> Uncertainty[float]:
        return self.support[0]

    @property
    def max_value(self) -> Uncertainty[float]:
        return self.support[1]

    @property
    def has_infinite_lower_support(self) -> bool:
        return self.min_value.value == -np.inf

    @property
    def has_infinite_upper_support(self) -> bool:
        return self.max_value.value == np.inf

    def make_uncertain(self) -> "Statistics":
        return Statistics(
            moments=[m.make_uncertain() for m in self.moments],
            support=(self.min_value.make_uncertain(), self.max_value.make_uncertain()),
        )

    def get(self, name: Literal["mean", "variance", "std_dev", "min_value", "max_value"], certain: bool = False) -> float:
        uncertainty = {
            "mean": self.mean,
            "variance": self.variance,
            "std_dev": self.std_dev,
            "min_value": self.min_value,
            "max_value": self.max_value,
        }[name]
        return uncertainty.get(name=name, certain=certain)

    def apply_algebraic_function(self, af: "AlgebraicFunction") -> "Statistics":
        scale = af.scale
        offset = af.offset
        new_min_max = (self.min_value * scale + offset, self.max_value * scale + offset)
        new_min, new_max = sort_uncertainties(new_min_max)

        mean = self.mean * scale + offset
        variance = self.variance * (scale**2)
        second_moment = variance + mean**2
        moments = [mean, second_moment]
        # TODO calculate higher order moments

        return Statistics(moments=moments, support=(new_min, new_max))


@runtime_checkable
class SupportsFloatAlgebra(Protocol):
    def __add__(self: T, x: float, /) -> T: ...
    def __radd__(self: T, x: float, /) -> T: ...
    def __sub__(self: T, x: float, /) -> T: ...
    def __rsub__(self: T, x: float, /) -> T: ...
    def __mul__(self: T, x: float, /) -> T: ...
    def __rmul__(self: T, x: float, /) -> T: ...
    def __truediv__(self: T, x: float, /) -> T: ...


F = TypeVar("F", bound=SupportsFloatAlgebra)


@dataclass(frozen=True)
class AlgebraicFunction:
    scale: float = 1.0
    offset: float = 0.0

    @cached_property
    def is_identity(self) -> bool:
        return self.scale == 1.0 and self.offset == 0.0

    @cached_property
    def is_active(self) -> bool:
        return not self.is_identity

    def __mul__(self, factor: float) -> "AlgebraicFunction":
        factor = float(factor)
        return AlgebraicFunction(scale=self.scale * factor, offset=self.offset * factor)

    def __rmul__(self, factor: float) -> "AlgebraicFunction":
        return self.__mul__(factor)

    def __add__(self, addend: float) -> "AlgebraicFunction":
        addend = float(addend)
        return AlgebraicFunction(scale=self.scale, offset=self.offset + addend)

    def __radd__(self, addend: float) -> "AlgebraicFunction":
        return self.__add__(addend)

    def apply(self, x: F) -> F:
        if self.is_identity:
            return x
        return self.scale * x + self.offset

    def apply_inverse(self, y: F) -> F:
        if self.is_identity:
            return y
        return (y - self.offset) / self.scale

    def apply_on_range(self, x_range: tuple[float, float]) -> tuple[float, float]:
        if self.is_identity:
            return x_range
        processed = self.apply(np.array(x_range))
        return float(np.min(processed)), float(np.max(processed))


@dataclass(frozen=True)
class ProbabilityDensityFunction:
    x: np.ndarray
    y: np.ndarray

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return np.interp(x, self.x, self.y, left=0.0, right=0.0)


@dataclass(frozen=True)
class ProbabilityMassFunction:
    x: np.ndarray
    y: np.ndarray

    def cdf(self, x: np.ndarray) -> np.ndarray:
        cumulative_probs = np.cumsum(self.y)
        result = np.empty_like(x, dtype=self.x.dtype)
        for i, xi in enumerate(x):
            idx = np.searchsorted(self.x, xi, side="right") - 1
            if idx < 0:
                result[i] = 0.0
            elif idx >= len(cumulative_probs):
                result[i] = 1.0
            else:
                result[i] = cumulative_probs[idx]
        return result

    def pmf(self, x: np.ndarray) -> np.ndarray:
        # For values with corresponding y, return y, else return 0.0
        return np.array([float(np.sum(self.y[self.x == xi])) for xi in x])

    def ppf(self, q: np.ndarray) -> np.ndarray:
        cumulative_probs = np.cumsum(self.y)
        result = np.empty_like(q, dtype=self.x.dtype)
        for i, quantile in enumerate(q):
            idx = np.searchsorted(cumulative_probs, quantile, side="left")
            result[i] = self.x[min(idx, len(self.x) - 1)]
        return result



from scipy.stats import beta, uniform

from randcraft.constructors import make_scipy
from randcraft.rvs.scipy_pdf import SciRV
from tests.base_test_case import BaseTestCase


class TestScipyRvs(BaseTestCase):
    def test_uniform_rv(self) -> None:
        rv = make_scipy(uniform, loc=1, scale=2)
        b = rv.scale(2)
        b = rv.scale(1)

        self.assertIsInstance(b._rv, SciRV)
        self.assertEqual(b.get_mean(), rv.get_mean())

    def test_beta_rv(self) -> None:
        rv = make_scipy(beta, 1, 2)
        b = rv.scale(1)
        self.assert_stats_are_close(a=b.statistics, b=rv.statistics)

        c = rv.scale(-1)
        self.assertEqual(c.get_mean(), rv.get_mean() * -1)
        self.assertEqual(c.get_variance(), rv.get_variance())



from abc import ABC, abstractmethod
from typing import Self

import numpy as np

from randcraft.models import AlgebraicFunction, ProbabilityDensityFunction, Statistics, Uncertainty
from randcraft.rvs.base import RV


class ContinuousRV(RV, ABC):
    @abstractmethod
    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction: ...

    @abstractmethod
    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]: ...

    def calculate_pmf(self) -> None:
        return None

    def _get_discrete_points(self) -> np.ndarray:
        points: list[float] = []
        if not self.statistics.has_infinite_lower_support:
            points.append(self.statistics.min_value.value)
        if not self.statistics.has_infinite_upper_support:
            points.append(self.statistics.max_value.value)
        return np.array(points)

    def scale(self, x: float) -> Self | "ScaledRV":
        # Override these if possible
        return ScaledRV(inner=self.copy(), algebraic_function=AlgebraicFunction(scale=x))

    def add_constant(self, x: float) -> Self | "ScaledRV":
        # Override these if possible
        return ScaledRV(inner=self.copy(), algebraic_function=AlgebraicFunction(offset=x))


class ScaledRV(ContinuousRV):  # TODO Add checks to cdf and ppf stuff
    def __init__(self, inner: ContinuousRV, algebraic_function: AlgebraicFunction) -> None:
        self._inner = inner
        self._af = algebraic_function
        assert self.algebraic_function.scale != 0.0, "Scale cannot be zero"

    @property
    def inner(self) -> ContinuousRV:
        return self._inner

    @property
    def algebraic_function(self) -> AlgebraicFunction:
        return self._af

    @property
    def short_name(self) -> str:
        return self.inner.short_name + "*"

    @property
    def statistics(self) -> Statistics:
        return self.inner.statistics.apply_algebraic_function(self.algebraic_function)

    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction:
        y = self.inner.calculate_pdf(self.algebraic_function.apply_inverse(x)).y / abs(self.algebraic_function.scale)
        return ProbabilityDensityFunction(x=x, y=y)

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self.inner.cdf(self.algebraic_function.apply_inverse(x))

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self.inner.ppf(x).apply(self.algebraic_function.apply)

    def scale(self, x: float) -> "ScaledRV":
        return ScaledRV(inner=self.inner, algebraic_function=self.algebraic_function * x)

    def add_constant(self, x: float) -> "ScaledRV":
        return ScaledRV(inner=self.inner, algebraic_function=self.algebraic_function + x)

    def sample_numpy(self, n: int) -> np.ndarray:
        return self.algebraic_function.apply(self.inner.sample_numpy(n))

    def _get_plot_range(self) -> tuple[float, float]:
        return self.algebraic_function.apply_on_range(self.inner._get_plot_range())

    def copy(self) -> "ScaledRV":
        return ScaledRV(inner=self.inner.copy(), algebraic_function=self.algebraic_function)



from typing import Any

import numpy as np


def clean_1d_array(x: Any | np.ndarray) -> list[float]:
    if isinstance(x, np.ndarray):
        assert x.ndim == 1, "Input numpy array must be 1D"
        return [float(v) for v in x.tolist()]
    return [float(v) for v in x]


def weighted_std(x: np.ndarray, weights: np.ndarray, unbiased: bool = True) -> float:
    average = np.average(x, weights=weights)
    variance = np.average((x - average) ** 2, weights=weights)
    if unbiased:
        eff_n = (weights.sum()) ** 2 / (np.sum(weights**2))
        if eff_n > 1:
            variance *= eff_n / (eff_n - 1)
    return np.sqrt(variance)



from collections.abc import Sequence
from functools import cached_property

import numpy as np
from scipy.signal import fftconvolve

from randcraft.models import ProbabilityDensityFunction, Statistics, Uncertainty, sum_uncertain_floats
from randcraft.rvs.base import CdfEstimator
from randcraft.rvs.continuous import ContinuousRV
from randcraft.rvs.discrete import DiracDeltaRV, DiscreteRV


class MultiRV(ContinuousRV):
    def __init__(
        self,
        continuous_pdfs: list[ContinuousRV],
        discrete_pdf: DiscreteRV = DiracDeltaRV(value=0.0),
    ) -> None:
        assert len(continuous_pdfs) > 0, "At least one continuous pdf is required"
        assert isinstance(discrete_pdf, DiscreteRV), f"discrete_pdf must be a {DiscreteRV.__name__}, got {type(discrete_pdf)}"
        self._continuous_pdfs = continuous_pdfs
        self._discrete_pdf = discrete_pdf

    @property
    def short_name(self) -> str:
        return "multi"

    @cached_property
    def pdfs(self) -> Sequence[ContinuousRV | DiscreteRV]:
        return self.continuous_pdfs + ([self._discrete_pdf] if self._discrete_pdf is not None else [])

    @property
    def continuous_pdfs(self) -> list[ContinuousRV]:
        return self._continuous_pdfs

    @cached_property
    def discrete_pdf(self) -> DiscreteRV:
        return self._discrete_pdf

    @cached_property
    def has_discrete_pdf(self) -> bool:
        if isinstance(self.discrete_pdf, DiracDeltaRV) and self.discrete_pdf.value == 0.0:
            return False
        return True

    @cached_property
    def statistics(self) -> Statistics:
        mean = sum_uncertain_floats(pdf.statistics.mean for pdf in self.pdfs)
        variance = sum_uncertain_floats(pdf.statistics.variance for pdf in self.pdfs)
        second_moment = mean**2 + variance
        # TODO calculate more moments

        min_value = sum_uncertain_floats([pdf.statistics.min_value for pdf in self.pdfs])
        max_value = sum_uncertain_floats([pdf.statistics.max_value for pdf in self.pdfs])

        return Statistics(
            moments=[mean, second_moment],
            support=(min_value, max_value),
        )

    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction:
        if not self.has_discrete_pdf:
            return ProbabilityDensityFunction(x=x, y=self._calculate_continuous_pdf(x))

        result = np.zeros_like(x)

        shifted_xs: list[np.ndarray] = []
        for offset in self.discrete_pdf.values:
            shifted_x = x - offset
            shifted_xs.append(shifted_x)

        # Find all unique values across all shifted_xs
        all_shifted_values = np.concatenate(shifted_xs)
        unique_shifted_values = np.unique(all_shifted_values)
        # Calculate continuous PDF only once for all unique shifted values
        unique_continuous = self._calculate_continuous_pdf(unique_shifted_values)
        # Create a mapping from value to index in unique_shifted_values
        idx_map = {val: idx for idx, val in enumerate(unique_shifted_values)}
        for offset, scale in zip(self.discrete_pdf.values, self.discrete_pdf.probabilities):
            shifted_x = x - offset
            # Map each value in shifted_x to its index in unique_shifted_values
            indices = np.array([idx_map.get(val, -1) for val in shifted_x])
            valid = indices >= 0
            result[valid] += scale * unique_continuous[indices[valid]]

        return ProbabilityDensityFunction(x=x, y=result)

    def _calculate_continuous_pdf(self, x: np.ndarray) -> np.ndarray:
        if len(self.continuous_pdfs) == 1:
            return self.continuous_pdfs[0].calculate_pdf(x).y

        assert x.ndim == 1, "Input numpy array must be 1D"
        low = np.min(x)
        high = np.max(x)
        spread = high - low
        start = low - spread
        end = high + spread
        x2 = np.linspace(start, end, len(x) * 3)
        raw_pdfs = [pdf.calculate_pdf(x2).y for pdf in self.continuous_pdfs]
        combined_pdf = fftconvolve(raw_pdfs[0], raw_pdfs[1], mode="same")
        for pdf in raw_pdfs[2:]:
            combined_pdf = fftconvolve(combined_pdf, pdf, mode="same")
        combined_pdf /= np.trapezoid(combined_pdf, x2)  # Normalize
        return np.interp(x, x2, combined_pdf)

    @cached_property
    def _cdf_estimator(self) -> CdfEstimator:
        return CdfEstimator(rv=self)

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._cdf_estimator.cdf(x)

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._cdf_estimator.ppf(x)

    def scale(self, x: float) -> "MultiRV":
        x = float(x)
        continuous_pdfs = [pdf.scale(x) for pdf in self.continuous_pdfs]
        discrete_pdf = self.discrete_pdf.scale(x)
        return MultiRV(continuous_pdfs=continuous_pdfs, discrete_pdf=discrete_pdf)

    def add_constant(self, x: float) -> "MultiRV":
        x = float(x)
        discrete_pdf = self.discrete_pdf.add_constant(x)
        return MultiRV(continuous_pdfs=self.continuous_pdfs, discrete_pdf=discrete_pdf)

    def sample_numpy(self, n: int) -> np.ndarray:
        return sum([pdf.sample_numpy(n) for pdf in self.pdfs])  # type: ignore

    def copy(self) -> "MultiRV":
        return MultiRV(continuous_pdfs=self.continuous_pdfs, discrete_pdf=self.discrete_pdf)



from unittest import TestCase

import numpy as np

from randcraft import make_dirac, make_discrete, make_gamma, make_normal, make_uniform
from randcraft.constructors import make_beta


class TestCoreRvs(TestCase):
    def test_normal_rv(self) -> None:
        mean = 23.4
        std_dev = 2.1

        rv = make_normal(mean=mean, std_dev=std_dev)

        sample = rv.sample_one()
        self.assertIsInstance(sample, float)

        self.assertEqual(rv.get_mean(exact=True), mean)
        self.assertAlmostEqual(rv.get_variance(exact=True), std_dev**2)
        self.assertAlmostEqual(rv.cdf(value=mean - std_dev, exact=True), 0.15865525393145707)
        self.assertAlmostEqual(rv.cdf(value=mean, exact=True), 0.5)
        self.assertAlmostEqual(rv.cdf(value=mean + std_dev, exact=True), 0.8413447460685429)

    def test_uniform_rv(self) -> None:
        low = 10.0
        high = 20.0

        rv = make_uniform(low=low, high=high)

        sample = rv.sample_one()
        self.assertIsInstance(sample, float)

        samples = rv.sample_numpy(n=100)
        self.assertIsInstance(samples, np.ndarray)
        self.assertEqual(samples.shape, (100,))
        self.assertGreaterEqual(min(samples), low)
        self.assertLessEqual(max(samples), high)

        self.assertEqual(rv.get_mean(exact=True), (low + high) / 2)
        self.assertAlmostEqual(rv.get_variance(exact=True), ((high - low) ** 2) / 12)
        self.assertAlmostEqual(rv.cdf(value=low, exact=True), 0.0)
        self.assertAlmostEqual(rv.cdf(value=15.0, exact=True), 0.5)
        self.assertAlmostEqual(rv.cdf(value=high, exact=True), 1.0)

    def test_beta_rv(self) -> None:
        a = 2.0
        b = 5.0

        rv = make_beta(a=a, b=b)

        sample = rv.sample_one()
        self.assertIsInstance(sample, float)
        self.assertGreaterEqual(sample, 0.0)
        self.assertLessEqual(sample, 1.0)

        samples = rv.sample_numpy(n=100)
        self.assertIsInstance(samples, np.ndarray)
        self.assertEqual(samples.shape, (100,))
        self.assertGreaterEqual(min(samples), 0.0)
        self.assertLessEqual(max(samples), 1.0)

        self.assertAlmostEqual(rv.get_mean(exact=True), a / (a + b))
        self.assertAlmostEqual(rv.get_variance(exact=True), (a * b) / ((a + b) ** 2 * (a + b + 1)))
        self.assertAlmostEqual(rv.cdf(value=0.0, exact=True), 0.0)
        self.assertAlmostEqual(rv.cdf(value=0.5, exact=True), 0.890625)
        self.assertAlmostEqual(rv.cdf(value=1.0, exact=True), 1.0)

    def test_gamma_rv(self) -> None:
        shape = 2.0
        scale = 3.0

        rv = make_gamma(a=shape, scale=scale)

        sample = rv.sample_one()
        self.assertIsInstance(sample, float)
        self.assertGreaterEqual(sample, 0.0)

        samples = rv.sample_numpy(n=100)
        self.assertIsInstance(samples, np.ndarray)
        self.assertEqual(samples.shape, (100,))
        self.assertGreaterEqual(min(samples), 0.0)

        self.assertAlmostEqual(rv.get_mean(exact=True), shape * scale)
        self.assertAlmostEqual(rv.get_variance(exact=True), shape * scale**2)
        self.assertAlmostEqual(rv.cdf(value=0.0, exact=True), 0.0)
        self.assertAlmostEqual(rv.cdf(value=shape * scale, exact=True), 0.5939941502901615)

    def test_dirac_delta_rv(self) -> None:
        value = 42.0

        rv = make_dirac(value=value)

        sample = rv.sample_one()
        self.assertEqual(sample, value)

        self.assertEqual(rv.get_mean(exact=True), value)
        self.assertAlmostEqual(rv.get_variance(exact=True), 0.0)
        self.assertEqual(rv.cdf(value=value - 1e-9, exact=True), 0.0)
        self.assertEqual(rv.cdf(value=value, exact=True), 1.0)
        self.assertEqual(rv.cdf(value=value + 1e-9, exact=True), 1.0)

    def test_discrete_rv(self) -> None:
        values = [1, 2, 3]
        probabilities = [0.2, 0.5, 0.3]

        rv = make_discrete(values=values, probabilities=probabilities)

        sample = rv.sample_numpy(n=100)
        self.assertIsInstance(sample, np.ndarray)
        self.assertEqual(sample.shape, (100,))
        self.assertTrue(np.all(np.isin(sample, values)))

        self.assertAlmostEqual(rv.get_mean(exact=True), sum(v * p for v, p in zip(values, probabilities)))
        self.assertAlmostEqual(
            rv.get_variance(exact=True),
            sum(p * (v - rv.get_mean(exact=True)) ** 2 for v, p in zip(values, probabilities)),
        )



from functools import cached_property

import numpy as np
from scipy.stats._distn_infrastructure import rv_continuous, rv_continuous_frozen

from randcraft.models import AlgebraicFunction, ProbabilityDensityFunction, Statistics, Uncertainty, certainly
from randcraft.rvs.continuous import ContinuousRV, ScaledRV


class RescalingError(Exception):
    pass


class SciRV(ContinuousRV):
    def __init__(self, scipy_rv_type: rv_continuous, *args, **kwargs) -> None:
        # It really should be continuous, but the typing here helps to interface with strange scipy typing
        scipy_rv = scipy_rv_type(*args, **kwargs)
        assert isinstance(scipy_rv, rv_continuous_frozen)
        self._scipy_rv_type = scipy_rv_type
        self._scipy_rv = scipy_rv

    @cached_property
    def short_name(self) -> str:
        try:
            name = self.scipy_rv.dist.name  # type: ignore
            return "scipy-" + name
        except AttributeError:
            return "scipy-unknown"

    @property
    def scipy_rv(self) -> rv_continuous_frozen:
        return self._scipy_rv

    @cached_property
    def statistics(self) -> Statistics:
        support = self.scipy_rv.support()
        lower = float(support[0])
        upper = float(support[1])

        return Statistics(
            moments=[certainly(self.scipy_rv.moment(n)) for n in range(1, 5)],
            support=(certainly(lower), certainly(upper)),
        )

    def scale(self, x: float) -> "SciRV | ScaledRV":
        return self._safe_rescale(AlgebraicFunction(scale=float(x), offset=0.0))

    def add_constant(self, x: float) -> "SciRV | ScaledRV":
        return self._safe_rescale(AlgebraicFunction(scale=1.0, offset=float(x)))

    def _safe_rescale(self, af: AlgebraicFunction) -> "SciRV | ScaledRV":
        def scale_with_scale_distribution() -> ScaledRV:
            return ScaledRV(inner=self, algebraic_function=af)

        has_finite_support_on_one_side = not (self.statistics.has_infinite_lower_support and self.statistics.has_infinite_upper_support)

        if has_finite_support_on_one_side and af.scale < 0:
            return scale_with_scale_distribution()

        def scale_with_scipy() -> SciRV:
            shape_args = self._scipy_rv.args
            shape_kwargs = {k: v for k, v in self._scipy_rv.kwds.items()}
            shape_kwargs["loc"] = 0.0
            shape_kwargs["scale"] = 1.0
            unit_distribution = self._scipy_rv_type(*shape_args, **shape_kwargs)  # type: ignore
            current_scale = self.std_dev / unit_distribution.std()
            current_loc = self.mean - unit_distribution.mean() * current_scale
            new_loc = af.apply(current_loc)
            new_scale = current_scale * af.scale
            shape_kwargs["loc"] = new_loc
            shape_kwargs["scale"] = new_scale
            return SciRV(self._scipy_rv_type, *shape_args, **shape_kwargs)

        for f in [scale_with_scipy, scale_with_scale_distribution]:
            result = f()
            expected_stats = self.statistics.apply_algebraic_function(af)
            result_stats = result.statistics

            all_close = True
            for m1, m2 in zip(expected_stats.moments, result_stats.moments):
                if not np.isclose(m1.value, m2.value):
                    all_close = False
                    break
            if not all_close:
                continue
            return result

        raise RescalingError(f"Could not rescale the {self._scipy_rv_type} correctly.")

    def sample_numpy(self, n: int) -> np.ndarray:
        return self.scipy_rv.rvs(size=n)

    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction:
        return ProbabilityDensityFunction(x=x, y=self.scipy_rv.pdf(x))

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly(self.scipy_rv.cdf(x))

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly(self.scipy_rv.ppf(x))

    def copy(self) -> "SciRV":
        return self.scale(1.0)  # type: ignore



from typing import TypeVar, Union, overload

import numpy as np

from randcraft.models import Statistics
from randcraft.pdf_convolver import PdfConvolver
from randcraft.rvs import (
    RV,
    DiracDeltaRV,
)
from randcraft.rvs.base import PdfPlotType

__all__ = ["RandomVariable"]

T = TypeVar("T")


class RandomVariable:
    # A wrapper around the different random variable implementations (RV)
    def __init__(self, rv: RV) -> None:
        self._rv = rv

    def __str__(self) -> str:
        mean = float(np.format_float_positional(x=self.statistics.mean.value, precision=3, fractional=False))
        var = float(np.format_float_positional(x=self.statistics.variance.value, precision=3, fractional=False))
        name = self._rv.short_name
        return f"<{self.__class__.__name__}({name}): {mean=}, {var=}>"

    def __repr__(self) -> str:
        return str(self)

    def __add__(self, other: Union["RandomVariable", float]) -> "RandomVariable":
        # Assumes pdfs are not correlated
        if not isinstance(other, RandomVariable):
            rv = DiracDeltaRV(value=float(other))
        else:
            rv = other._rv
        new_rv = PdfConvolver.convolve_pdfs(pdfs=[self._rv, rv])  # type: ignore
        return RandomVariable(rv=new_rv)

    def __radd__(self, other: Union["RandomVariable", float]) -> "RandomVariable":
        return self.__add__(other)

    def __sub__(self, other: Union["RandomVariable", float]) -> "RandomVariable":
        # Assumes pdfs are not correlated
        if isinstance(other, RandomVariable):
            return self + other.scale(-1)
        else:
            return self + (float(other) * -1)

    @property
    def statistics(self) -> Statistics:
        return self._rv.statistics

    def get_mean(self, exact: bool = False) -> float:
        return self._rv.statistics.get(name="mean", certain=exact)

    def get_variance(self, exact: bool = False) -> float:
        return self._rv.statistics.get(name="variance", certain=exact)

    def get_min_value(self, exact: bool = False) -> float:
        return self._rv.statistics.get(name="min_value", certain=exact)

    def get_max_value(self, exact: bool = False) -> float:
        return self._rv.statistics.get(name="max_value", certain=exact)

    @overload
    def cdf(self, value: float | int, exact: bool = False) -> float: ...

    @overload
    def cdf(self, value: np.ndarray, exact: bool = False) -> np.ndarray: ...

    def cdf(self, value: np.ndarray | float | int, exact: bool = False) -> float | np.ndarray:
        if isinstance(value, np.ndarray):
            return self._rv.cdf(x=value).get(name="cdf", certain=exact)
        return self.cdf(value=np.array([float(value)]))[0]

    @overload
    def ppf(self, value: float | int, exact: bool = False) -> float: ...

    @overload
    def ppf(self, value: np.ndarray, exact: bool = False) -> np.ndarray: ...

    def ppf(self, value: np.ndarray | float | int, exact: bool = False) -> float | np.ndarray:
        if isinstance(value, np.ndarray):
            return self._rv.ppf(q=value).get(name="ppf", certain=exact)
        return self.ppf(value=np.array([value]))[0]

    def sample_numpy(self, n: int) -> np.ndarray:
        return self._rv.sample_numpy(n=n)

    def sample_one(self) -> float:
        return self.sample_numpy(1).tolist()[0]

    def scale(self, factor: float) -> "RandomVariable":
        if factor == 0.0:
            return RandomVariable(DiracDeltaRV(value=0.0))
        factor = float(factor)
        new_pdf = self._rv.scale(x=factor)
        return RandomVariable(rv=new_pdf)

    def multi_sample(self, n: int) -> "RandomVariable":
        # Return an RV that is the sum of n independent samples of this RV
        if n <= 0:
            raise ValueError("n must be a positive integer.")
        if n == 1:
            return self
        new_rv = PdfConvolver.convolve_pdfs(pdfs=[self._rv] * n)  # type: ignore
        return RandomVariable(rv=new_rv)

    def plot(self, kind: PdfPlotType = "both") -> None:
        self._rv.plot(kind=kind)



import numpy as np

from randcraft.anonymous import make_anonymous_continuous_rv_from_sampler
from tests.base_test_case import BaseTestCase


class TestAnon(BaseTestCase):
    def test_anonymous_rv(self) -> None:
        print("hello")

        def sampler(n: int) -> np.ndarray:
            generator = np.random.default_rng()
            return 20.0 + generator.random(n)

        rv = make_anonymous_continuous_rv_from_sampler(sampler=sampler)

        # Statistics are not exact
        print(rv.get_mean())
        self.assertTrue(20.0 <= rv.get_mean() <= 21.0)
        self.assertTrue(0.0 < rv.get_variance() <= 1.0)
        print("hi")
        self.assertLess(rv.cdf(value=20.0), 0.1)
        self.assertGreater(rv.cdf(value=21.0), 0.9)

        self.assertRaises(NotImplementedError, lambda: rv.get_mean(exact=True))
        self.assertRaises(NotImplementedError, lambda: rv.get_variance(exact=True))



import logging
from collections.abc import Callable

import numpy as np
from scipy.stats import gaussian_kde

from randcraft.constructors import make_discrete
from randcraft.random_variable import RandomVariable
from randcraft.rvs.discrete import DiscreteRV
from randcraft.rvs.gaussian_kde import GaussianKdeRV

logger = logging.getLogger(__name__)


def make_gaussian_kde(
    observations: np.ndarray | RandomVariable,
    bw_method: str | float | Callable | None = None,
    weights: np.ndarray | None = None,
) -> RandomVariable:
    """
    Using a set of observations create a gaussian kernel density estimate (KDE) using scipy.stats.gaussian_kde

    Args:
        observations (np.ndarray): Either a 1D numpy array of observations or a RandomVariable containing a discrete distribution of observations.
        bw_method (str | float | Callable | None, optional): The bandwidth selector for the scipy gaussian_kde method. Defaults to None.
        weights (np.ndarray | None, optional): The weights to use for the KDE. Defaults to None.

    Returns:
        RandomVariable: A random variable with a continuous distribution representing the KDE.
    """
    if not isinstance(observations, (np.ndarray, RandomVariable)):
        raise TypeError("Observations must be either a 1D numpy array or a RandomVariable")

    if isinstance(observations, RandomVariable):
        discrete_rv = observations
        assert isinstance(discrete_rv._rv, DiscreteRV), "RandomVariable for observations must have a discrete distribution"
        np_observations = np.array(discrete_rv._rv.values)
    else:
        discrete_rv = make_discrete_rv_from_observations(observations=observations, weights=weights, reduce=True)
        np_observations = observations

    discerete_inner_rv = discrete_rv._rv
    assert isinstance(discerete_inner_rv, DiscreteRV)
    kde_model = gaussian_kde(dataset=np_observations, bw_method=bw_method, weights=weights)
    return RandomVariable(rv=GaussianKdeRV(discrete=discerete_inner_rv, kde=kde_model))


def reduce_observations(observations: np.ndarray, weights: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray | None]:
    """Reduce observations by grouping identical values and summing their weights."""
    assert len(observations) > 1, "Need multiple observations for KDE"
    assert observations.ndim == 1, "Input data must be a 1D numpy array"
    if weights is not None:
        unique, inverse = np.unique(observations, return_inverse=True)
        summed_weights = np.zeros_like(unique, dtype=float)
        for idx, w in zip(inverse, weights):
            summed_weights[idx] += w
        return unique, summed_weights
    else:
        unique = np.unique(observations)
        return unique, None


def make_discrete_rv_from_observations(observations: np.ndarray, weights: np.ndarray | None = None, reduce: bool = True) -> RandomVariable:
    if reduce:
        observations, weights = reduce_observations(observations=observations, weights=weights)
    return make_discrete(values=observations, probabilities=weights)



from abc import ABC, abstractmethod
from typing import Literal, Self, TypeVar

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from scipy.integrate import cumulative_trapezoid

from randcraft.models import ProbabilityDensityFunction, ProbabilityMassFunction, Statistics, Uncertainty, maybe

type PdfPlotType = Literal["pdf", "cdf", "both"]


class RV(ABC):
    @property
    @abstractmethod
    def short_name(self) -> str: ...

    @property
    @abstractmethod
    def statistics(self) -> Statistics: ...

    @abstractmethod
    def sample_numpy(self, n: int) -> np.ndarray: ...

    @abstractmethod
    def scale(self, x: float) -> "RV": ...

    @abstractmethod
    def add_constant(self, x: float) -> "RV": ...

    @abstractmethod
    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction | None: ...

    @abstractmethod
    def calculate_pmf(self) -> ProbabilityMassFunction | None: ...

    @abstractmethod
    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]: ...

    @abstractmethod
    def ppf(self, q: np.ndarray) -> Uncertainty[np.ndarray]: ...

    @abstractmethod
    def _get_discrete_points(self) -> np.ndarray: ...

    @abstractmethod
    def copy(self) -> Self: ...

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}(mean={self.mean}, variance={self.variance})>"

    def __repr__(self) -> str:
        return str(self)

    @property
    def stats(self) -> Statistics:
        return self.statistics

    @property
    def mean(self) -> float:
        return self.stats.mean.value

    @property
    def variance(self) -> float:
        return self.stats.variance.value

    @property
    def std_dev(self) -> float:
        return self.stats.std_dev.value

    @property
    def min_value(self) -> float:
        return self.stats.min_value.value

    @property
    def max_value(self) -> float:
        return self.stats.max_value.value

    def plot(self, kind: PdfPlotType = "both") -> None:
        start, end = self._get_plot_range()
        discrete_points = self._get_discrete_points()
        x = np.linspace(start, end, 1000)
        if len(discrete_points) > 0:
            step = (x[1] - x[0]) / 100  # TODO Use double index instead of small step
            x = np.unique(np.concatenate([x, discrete_points - step, discrete_points]))
            x.sort()

        v_lines = [(self.mean, "red", "mean")]
        if self.variance > 0:
            v_lines.append((self.mean - self.std_dev, "orange", "-1 std_dev"))
            v_lines.append((self.mean + self.std_dev, "orange", "+1 std_dev"))

        def _plot(is_cumulative: bool, ax: Axes) -> None:
            if is_cumulative:
                self.plot_cdf_on_axis(ax=ax, x=x)
                ax.set_title("CDF")
                ax.set_ylim(0.0, 1.01)
            else:
                self.plot_pdf_on_axis(ax=ax, x=x)
                ax.set_title("PDF")
                ax.set_ylim(bottom=0)
            ax.set_xlabel("x")
            ax.set_xlim(self._get_plot_range())
            ax.set_ylabel("P(X<=x)" if is_cumulative else "P(X=x)")
            for item in v_lines:
                pos, color, label = item
                ax.axvline(pos, color=color, label=label, linestyle="--", linewidth=1)
            ax.legend()
            ax.grid(True)

        if kind == "both":
            fig, axs = plt.subplots(2, 1, sharex="all")
            _plot(is_cumulative=False, ax=axs[0])
            _plot(is_cumulative=True, ax=axs[1])
        elif kind == "pdf":
            fig, ax1 = plt.subplots()
            _plot(is_cumulative=False, ax=ax1)
        elif kind == "cdf":
            fig, ax1 = plt.subplots()
            _plot(is_cumulative=True, ax=ax1)
        else:
            raise ValueError(f"Invalid kind: {kind}. Choose 'pdf', 'cdf', or 'both'.")

        plt.tight_layout()
        fig.set_size_inches(10, 6)
        plt.show()

    def _get_plot_range(self) -> tuple[float, float]:
        if not np.isinf(self.min_value):
            start = self.min_value - 0.1 * self.std_dev
        else:
            start = self.mean - 4 * self.std_dev
        if not np.isinf(self.max_value):
            end = self.max_value + 0.1 * self.std_dev
        else:
            end = self.mean + 4 * self.std_dev
        buffer = (end - start) * 0.01
        if buffer == 0.0:
            buffer = max(1.0, abs(self.mean))
        return start - buffer, end + buffer

    def plot_pdf_on_axis(self, ax: Axes, x: np.ndarray) -> None:
        cont_pdf = self.calculate_pdf(x)
        if cont_pdf is not None:
            ax.plot(cont_pdf.x, cont_pdf.y)

        disc_pdf = self.calculate_pmf()
        if disc_pdf is not None:
            for x, p in zip(disc_pdf.x, disc_pdf.y):
                ax.vlines(x, 0, p, colors="C0", linewidth=2)
                ax.scatter(x, p, color="C0", s=50, zorder=5)
        return

    def plot_cdf_on_axis(self, ax: Axes, x: np.ndarray) -> None:
        y = self.cdf(x).value
        ax.plot(x, y)


T_RV = TypeVar("T_RV", bound=RV)


class CdfEstimator:
    def __init__(self, rv: RV) -> None:
        self.rv = rv
        self.x, self.p = self.calculate_cdf(rv)

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return maybe(np.interp(x, self.x, self.p))

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return maybe(np.interp(x, self.p, self.x))

    @staticmethod
    def calculate_cdf(rv: RV) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns two numpy arrays (x_values, cumulative_probabilities) representing the CDF.
        The chance that x < value can be found by interpolating cumulative_probabilities at value.
        """
        mean = rv.statistics.mean.value
        std_dev = rv.statistics.std_dev.value
        has_finite_lower_support = not np.isinf(rv.statistics.min_value.value)
        has_finite_upper_support = not np.isinf(rv.statistics.max_value.value)

        if has_finite_lower_support and has_finite_upper_support:
            lower = rv.statistics.min_value.value
            upper = rv.statistics.max_value.value
        elif has_finite_lower_support:
            lower = rv.statistics.min_value.value
            upper = lower + 6 * std_dev
        elif has_finite_upper_support:
            upper = rv.statistics.max_value.value
            lower = upper - 6 * std_dev
        else:
            lower = mean - 3 * std_dev
            upper = mean + 3 * std_dev

        x_values = np.linspace(lower, upper, 10000)
        pdf = rv.calculate_pdf(x_values)
        assert pdf is not None, "PDF must be defined to calculate CDF"
        pdf_vals = pdf.y

        cdf_vals = cumulative_trapezoid(pdf_vals, x_values, initial=0)
        if cdf_vals[-1] < 1:
            remainder = 1 - cdf_vals[-1]
            cdf_vals = cdf_vals + remainder / 2
            x_values = np.concatenate(([-np.inf], x_values, [np.inf]))
            cdf_vals = np.concatenate(([0.0], cdf_vals, [1.0]))
        else:
            cdf_vals /= cdf_vals[-1]  # Normalize
            cdf_vals[0] = 0.0
            cdf_vals[-1] = 1.0

        return x_values, cdf_vals



from functools import cached_property

import numpy as np

from randcraft.models import ProbabilityMassFunction, Statistics, Uncertainty, certainly
from randcraft.rvs.base import RV


class DiscreteRV(RV):
    def __init__(self, values: list[float], probabilities: list[float] | None = None) -> None:
        if probabilities is None:
            probabilities = [1.0 / len(values)] * len(values)
        assert len(values) == len(probabilities), "Values and probabilities must have the same length."
        assert all(p >= 0 for p in probabilities), "Probabilities must be non-negative."
        assert abs(sum(probabilities) - 1.0) < 1e-6, "Probabilities must sum to 1."
        assert len(set(values)) == len(values), "Values must be unique."

        self._values = np.array(values)
        self._probabilities = np.array(probabilities)

    @property
    def short_name(self) -> str:
        return "discrete"

    @cached_property
    def statistics(self) -> Statistics:
        moments = [certainly(np.sum(self._values ** (k + 1) * self._probabilities)) for k in range(4)]
        support = (certainly(float(np.min(self._values))), certainly(float(np.max(self._values))))
        return Statistics(moments=moments, support=support)

    @cached_property
    def values(self) -> list[float]:
        return self._values.tolist()

    @cached_property
    def probabilities(self) -> list[float]:
        return self._probabilities.tolist()

    @cached_property
    def _pmf(self) -> ProbabilityMassFunction:
        return self.calculate_pmf()

    def pmf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly(self._pmf.pmf(x))

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly(self._pmf.cdf(x))

    def ppf(self, q: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly(self._pmf.ppf(q))

    def _get_discrete_points(self) -> np.ndarray:
        return np.array(self.values)

    def calculate_pdf(self, x: np.ndarray) -> None:
        return None

    def calculate_pmf(self) -> ProbabilityMassFunction:
        return ProbabilityMassFunction(x=self._values, y=self._probabilities)

    def scale(self, x: float) -> "DiscreteRV":
        x = float(x)
        return DiscreteRV(values=[float(v * x) for v in self._values], probabilities=self._probabilities.tolist())

    def add_constant(self, x: float) -> "DiscreteRV":
        x = float(x)
        return DiscreteRV(values=[float(v + x) for v in self._values], probabilities=self._probabilities.tolist())

    def sample_numpy(self, n: int) -> np.ndarray:
        rng = np.random.default_rng()
        return rng.choice(self._values, size=n, p=self._probabilities)

    def copy(self) -> "DiscreteRV":
        return DiscreteRV(values=self.values, probabilities=self.probabilities)


class DiracDeltaRV(DiscreteRV):
    def __init__(self, value: float) -> None:
        super().__init__(values=[value])

    @property
    def short_name(self) -> str:
        return "dirac"

    @cached_property
    def statistics(self) -> Statistics:
        moments = [certainly(self.value ** (k + 1)) for k in range(4)]
        support = (certainly(float(self.value)), certainly(float(self.value)))
        return Statistics(moments=moments, support=support)

    @cached_property
    def value(self) -> float:
        return float(self._values[0])

    def pmf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly((x == self.value).astype(float) * 1.0)

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly((x >= self.value).astype(float) * 1.0)

    def ppf(self, q: np.ndarray) -> Uncertainty[np.ndarray]:
        return certainly(np.ones_like(q) * self.value)

    def scale(self, x: float) -> "DiracDeltaRV":
        return DiracDeltaRV(value=self.mean * x)

    def add_constant(self, x: float) -> DiscreteRV:
        return DiracDeltaRV(value=self.mean + x)

    def sample_numpy(self, n: int) -> np.ndarray:
        return np.ones(n) * self.value

    def copy(self) -> "DiracDeltaRV":
        return DiracDeltaRV(value=self.value)



import logging
from collections.abc import Callable

import numpy as np

from randcraft.models import ProbabilityDensityFunction, Statistics, Uncertainty
from randcraft.rvs.continuous import ContinuousRV

logger = logging.getLogger(__name__)


class AnonymousRV(ContinuousRV):
    def __init__(
        self,
        sampler: Callable[[int], np.ndarray],  # Function used for sampling
        ref_rv: ContinuousRV,  # Reference used for calculating statistics, pdf, cdf etc
        external_statistics: Statistics | None = None,  # Optionally override statistics
    ) -> None:
        self._sampler = sampler
        self._ref_rv = ref_rv
        self._statistics = (external_statistics or ref_rv.statistics).make_uncertain()

    @property
    def short_name(self) -> str:
        return "anon"

    @property
    def statistics(self) -> Statistics:
        return self._statistics

    def sample_numpy(self, n: int) -> np.ndarray:
        return self._ref_rv.sample_numpy(n)

    def ppf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._ref_rv.ppf(x).make_uncertain()

    def cdf(self, x: np.ndarray) -> Uncertainty[np.ndarray]:
        return self._ref_rv.cdf(x).make_uncertain()

    def calculate_pdf(self, x: np.ndarray) -> ProbabilityDensityFunction:
        return self._ref_rv.calculate_pdf(x)

    def copy(self) -> "AnonymousRV":
        return AnonymousRV(sampler=self._sampler, ref_rv=self._ref_rv)



"""
randcraft
"""

from randcraft.constructors import (
    make_coin_flip,
    make_die_roll,
    make_dirac,
    make_discrete,
    make_gamma,
    make_normal,
    make_uniform,
)

__all__ = [
    "make_normal",
    "make_uniform",
    "make_discrete",
    "make_dirac",
    "make_coin_flip",
    "make_die_roll",
    "make_gamma",
]
